import os.path, sys
from PIL import Image

FileExtension = ".png"
OutputName = "BITMAPS"
MaxImages = 63
MaxWidthP = 256
MaxHeightP = 192

FileNumber = 0
ImageCount = 0

BlackRGB = (  0,   0,   0)
WhiteRGB = (255, 255, 255)
KeyRGB   = (255,   0, 255)  # Magenta.
GreyRGB  = (128, 128, 128)  # May be used later for "auto hatching."

BitmapData = bytearray()
BitmapIndex = ""
BitmapOffset = 0

CompressData = True
InterleaveAlpha = True

# Not used, but kept for future reference.
def CompressImageRLE(Data, ImageNum, GetSizeOnly):
    CompressedImage = bytearray()    
    MaxRunCount = 255
    CompressedSize = 0
    PreviousByte = None
    CurrentByte = None
    RunCount = 0
    MidRun = False
    

    for Byte in Data:
        PreviousByte = CurrentByte
        CurrentByte = Byte
        
        # Use RLE when matching pair found.
        if (PreviousByte == CurrentByte):
            if MidRun:
                RunCount +=1
                # Check against run length encoding limit.
                if (RunCount == MaxRunCount):
                    # Store run length.
                    CompressedImage.extend(RunCount.to_bytes(1, byteorder="big"))
                    # Break from run.
                    PreviousByte = None
                    MidRun = False
            else:
                # Store second byte of pair.
                CompressedImage.extend(Byte.to_bytes(1, byteorder="big"))
                # Start counting run.
                MidRun = True
                RunCount = 0
                
        # Otherwise, break from any current run and store raw byte.
        else:
            if MidRun:
                # Store run length.
                CompressedImage.extend(RunCount.to_bytes(1, byteorder="big"))
                # Break from run.
                PreviousByte = None
                MidRun = False
            # Store raw byte.
            CompressedImage.extend(Byte.to_bytes(1, byteorder="big"))
            
    if GetSizeOnly:
        print("Uncompressed image", ImageNum, "is", len(Data), "bytes.")
        print("RLE compressed image would be", len(CompressedImage), "bytes.")
        return
    else:
        return CompressedImage

def HasAlphaKey(Image):    
    ImageRGB = Image.convert()
    RawDump = list(ImageRGB.getdata())

    AlphaKey=False
    for p in RawDump:
        if p==KeyRGB:
            AlphaKey=True
            break

    return AlphaKey

def GetMapLength(Image):
    if HasAlphaKey(Image):
        MapPitch = Image.width / (2*8)
    else:
        MapPitch = Image.width / (4*8)
    # Adjust for part-filled  bytes.
    if MapPitch.is_integer():
        MapPitch = int(MapPitch)
    else:
        MapPitch = int(MapPitch)
        MapPitch += 1
    MapLength = int(MapPitch * CurrentImage.height / 8)
    return MapLength

# Returns code for the CellMap and a zero/two/eight byte array for the PixelData.
def ConvertCell(CellData, CurrentImageData, CurrentImageStart):
    global BitmapData
    global BitmapOffset
    
    CellSum = sum(CellData)
    DataBlock = bytearray()

    if CellSum == 0:
        # WhiteBlock (8*0.)
        MapCode = 0b10
    elif CellSum == 2040:
        # BlackBlock (8*255.)
        MapCode = 0b11        
    else:
        # Initialise CellPointer (may be no compiled data yet.)
        CellPointer = None
        # Search compiled data from previous images.
        for offset in range(0, len(BitmapData)-7 ):        
            CellPointer = True
            for byte in range(8):
                if BitmapData[offset+byte] != CellData[byte]:
                    CellPointer = None
                    break
            if CellPointer:
                CellPointer = (BitmapOffset-len(BitmapData)) + offset
                break                
        # No match from previous compiled data?
        if CellPointer == None:
            # Search current image data.
            for offset in range(0, len(CurrentImageData)-7 ):        
                CellPointer = True
                for byte in range(8):
                    if CurrentImageData[offset+byte] != CellData[byte]:
                        CellPointer = None
                        break
                if CellPointer:
                    CellPointer = CurrentImageStart + offset
                    break
            
        if CellPointer == None:
            MapCode = 0b00
            DataBlock.extend(CellData)
        else:
            MapCode = 0b01
            DataBlock.extend(CellPointer.to_bytes(2, byteorder="little"))
            
    return(MapCode, DataBlock)
        

# Cell-based compression technique.
def CompressImageBlock(Data, ImageNum, ImageWidth, ImageHeight, CompressPixelData, InterleaveAlpha, CurrentImageStart):
    CompressedImage = bytearray()
    MapData = bytearray()
    PixelData = bytearray()

    # MapData is two bits per cell, so four items per byte.
    MapPackByte = 0b00000000
    MapPackLimit = 4
    MapPackStep = 0

    # CellDim may cause odd behaviour if changed from 8.
    CellDim = 8
    CellsWide = int(ImageWidth / CellDim)
    CellsHigh = int(ImageHeight / CellDim)

    if InterleaveAlpha:
        MonoOffset = int(len(Data)/2)

    for row in range(CellsHigh):
        RowData = bytearray()
        
        for column in range (CellsWide):
            CellData = bytearray()
            CellOffset = column + (row * ImageWidth)

            for byte in range (CellDim):
                CellData.append(Data[CellOffset])
                CellOffset += CellsWide
                
            # Mono channel.
            ConvertedCell = ConvertCell(CellData, PixelData, CurrentImageStart)

            # Append CellCode to current MapData byte.
            # Part MapData bytes will contain 0b00 in unused areas.
            # Pack cell code into upper two bits of byte:
            # - multiplying by 2^6 shifts them six places right to upper two bits.
            MapPackByte += (ConvertedCell[0] * 64)
            MapPackStep +=1
            
            # Extend PixelData by pointer or pixel bytes, where applicable.
            if len(ConvertedCell[1]) > 0:
                RowData.extend(ConvertedCell[1])

            # Repeat above for Mono data after Alpha, if interleaving.
            if InterleaveAlpha:
                CellData = bytearray()
                CellOffset = column + (row * ImageWidth)
                CellOffset += MonoOffset
                for byte in range (CellDim):
                   CellData.append(Data[CellOffset])
                   CellOffset += CellsWide
                ConvertedCell = ConvertCell(CellData, PixelData, CurrentImageStart)
                        
                # Shift bits in byte two places right (divide by four.)
                MapPackByte = int(MapPackByte/4)
                # No risk of overflow, as MapPackLimit is multiple of two.
                MapPackByte += (ConvertedCell[0] * 64)
                MapPackStep +=1
                
                if len(ConvertedCell[1]) > 0:
                    RowData.extend(ConvertedCell[1])
                    
            # MapData packing is the same, regardless of interleave.
            if MapPackStep==MapPackLimit:
                MapData.append(MapPackByte)
                MapPackByte = 0b00000000
                MapPackStep = 0
            elif column+1==CellsWide:
                # Move bits to lower end of any part-filled byte.
                for shift in range(MapPackStep, MapPackLimit):
                   MapPackByte = int(MapPackByte/4)
                MapData.append(MapPackByte)
                MapPackByte = 0b00000000
                MapPackStep = 0 
            else:
                # Shift bits in byte two places right (divide by four.)
                MapPackByte = int(MapPackByte/4)

            # Pack pixel data only at end of each row.
            if column+1 == CellsWide:
                if CompressPixelData:
                    RowData = CompressImageRLE(RowData, 0, False)
                PixelData.extend(RowData)

    # Join PixelData after MapData in order to create CompressedImage.
    CompressedImage.extend(MapData)
    CompressedImage.extend(PixelData)
    print ("Image", ImageNum, "block compressed from",
           len(Data), "to", len(CompressedImage), "bytes.")    
    return CompressedImage

def ConvertImage(Image, ImageNum, CompressData, InterleaveAlpha):
    global BitmapData
    global BitmapIndex
    global BitmapOffset
    global MaxWidthP
    global MaxHeightP
    
    ImageRGB = Image.convert()
    RawDump = list(ImageRGB.getdata())
    
    Width = ImageRGB.width
    Height = ImageRGB.height
    Size = Width * Height
    AlphaLayer = False
    AlphaChannel = False
    MonoChannel = False
    AlphaByte = 0
    MonoByte = 0
    BitCount = 0
    AlphaData = bytearray()
    MonoData = bytearray()
    CompiledData = bytearray()

    CurrentImageStart = SpecDataStart + len(BitmapData) + GetMapLength(Image)
    
    if (Width%8 > 0) or (Height%8 > 0):
        print("Error at Image", ImageNum)
        print("Dimensions", CurrentImage.size, "are not both multiples of 8.")
        sys.exit(1)

    if (Width > MaxWidthP) or (Height > MaxHeightP):
        print("Error at Image", ImageNum)
        print("Dimensions", CurrentImage.size, "exceed maximum width and/or height.")
        sys.exit(1)
        
        
    for p in range(Size):
        AlphaByte += AlphaByte
        MonoByte += MonoByte
        Pixel = RawDump[p]
        
        if Pixel == KeyRGB:
            AlphaLayer = True
            AlphaChannel = True
            MonoChannel = False
        elif Pixel == WhiteRGB:
            MonoChannel = False
            AlphaChannel = False
        elif Pixel == BlackRGB:
            MonoChannel = True
            AlphaChannel = False
        else:
            print("Error at Image", ImageNum)
            print("Invalid pixel colour", Pixel)
            sys.exit(1)

        if AlphaChannel:
            AlphaByte +=1
        if MonoChannel:
            MonoByte +=1

        BitCount +=1
        if BitCount == 8:
            BitCount = 0
            AlphaData.extend(AlphaByte.to_bytes(1, byteorder="big"))
            AlphaByte = 0
            MonoData.extend(MonoByte.to_bytes(1, byteorder="big"))
            MonoByte = 0
            
    # Data bits are packed, now sequence and compress by flags.    
    if CompressData:
        if AlphaLayer:
            if InterleaveAlpha:
                CompiledData.extend(AlphaData)
                CompiledData.extend(MonoData)
                CompiledData = CompressImageBlock(CompiledData, ImageNum, Width, Height, False, True, CurrentImageStart)
            else:
                CompiledData.extend(CompressImageBlock(AlphaData, ImageNum, Width, Height, False, False, CurrentImageStart))               
                CompiledData.extend(CompressImageBlock(MonoData, ImageNum, Width, Height, False, False, CurrentImageStart))
        else:
            CompiledData.extend(CompressImageBlock(MonoData, ImageNum, Width, Height, False, False, CurrentImageStart))
    else:
        if AlphaLayer:
            if InterleaveAlpha:
                for i in range(len(MonoData)):
                    CompiledData.append(AlphaData[i])
                    CompiledData.append(MonoData[i])
            else:
                CompiledData.extend(AlphaData)
                CompiledData.extend(MonoData)
        else:
            CompiledData.extend(MonoData)

    # Add sequenced image data to set.
    BitmapData.extend(CompiledData)
    
    # Now need to add information to text index and binary data array.
    Width = int(Width / 8)
    Height = int(Height / 8)
    if AlphaLayer:
        # Bit 7 of Width is AlphaLayer flag.
        Width = Width + 128
    if CompressData:
        # Bit 6 of Width is Compressed flag.
        Width = Width + 64
    BitmapIndex +=("  DW " + str(BitmapOffset) + "\n")
    BitmapIndex +=("  DB " + str(Width) + ", " + str(Height) + "\n")
    BitmapOffset += len(CompiledData)
        
    return True

# Main entry.
SpecDataStart = input("Please enter the Bitmap data hex start address:")
SpecDataStart = int(SpecDataStart, 16)

if SpecDataStart < 0 or SpecDataStart > 65535:
    print("Error: data start address", SpecDataStart, "is outside memory range.")
    sys.exit(1)

BitmapOffset = SpecDataStart

for FileNumber in range(MaxImages):
    FileNumString = str(FileNumber).zfill(3) + FileExtension
    if os.path.isfile(FileNumString):
        # Add the data.
        CurrentImage = Image.open(FileNumString)
        ConvertImage(CurrentImage, FileNumber, CompressData, InterleaveAlpha)
        CurrentImage.close()
        ImageCount +=1
    else:
        # Add an empty entry.
        BitmapIndex +=("  DW " + str(BitmapOffset) + "\n")
        BitmapIndex +=("  DB 0, 0 \n")
        print("Null entry for image", FileNumber)

print("\nFound", ImageCount, "image files.")
print("Generated", len(BitmapData), "bytes of data.")

if (SpecDataStart+len(BitmapData) > 65535):
    print("Error: data too large for space following start address.")
    sys.exit(1)

IndexOut = open("BITMAPS-I.asm", "w")
IndexOut.write(BitmapIndex)
IndexOut.close()

DataOut = open("BITMAPS.bin", "wb")
DataOut.write(BitmapData)
DataOut.close()

print("Index and Data files written.")
sys.exit(0)
