# Simple version, no support for FLASH attributes.
import os.path, sys
from PIL import Image

gFileInExt = ".scr"
gFileOutExt = ".bin"
gAsmOutExt = ".asm"
gImgOutExt = ".png"
gIndexExt = "-I"
gAttrFile = "ATTRS"

gMaxFiles = 127
gScrLength = 6912
gScrAttStart = 6144
gAttrOffset = int(0xDF00)
gScrX = 256
gScrY = 192


def PixelsToBmp(Data, FileName):
    pitch = int(gScrX/8)
    bmp = Image.new("RGB",(gScrX, gScrY))
    for row in range(gScrY):
        # Must translate .scr pixel order from ZX screen layout to linear format.
        yCell = int((row-row%8)/8)
        sourceStart = 32 * (yCell%8)
        sourceStart += 256 * (row%8)
        sourceStart += 2048 * (int((yCell-yCell%8)/8))
        #print(yCell, sourceStart)
        for column in range(pitch):            
            sourceOffset = column
            byte = Data[sourceStart+column]
            for bit in range(8):
                if ((byte<<bit) & 0b10000000):
                    pixValue = (0, 0, 0)
                else:
                    pixValue = (255, 255, 255)
                bmp.putpixel(((column*8)+bit, row), pixValue)
                
    bmp.save(FileName)
    return
    

def CompressAttributes(Attrs):
    RunCount = 0
    RunLimit = 127
    CurrentByte = bytes(1)
    CompAttrs = bytearray()
    
    CompAttrs.append(Attrs[0])
    for i in range(1,len(Attrs)):
        CurrentByte = Attrs[i]
        if CurrentByte == Attrs[i-1]:
            if RunCount < RunLimit:
                RunCount += 1
            else:
                RunCount += 128
                CompAttrs.append(RunCount)
                CompAttrs.append(CurrentByte)
                RunCount = 0
        else:
            if RunCount == 0:
                CompAttrs.append(CurrentByte)
            else:
                RunCount += 128
                CompAttrs.append(RunCount)
                CompAttrs.append(CurrentByte)
                RunCount = 0
    if RunCount > 0:
        RunCount += 128
        CompAttrs.append(RunCount)        
    return CompAttrs

index = gAttrFile + gIndexExt + gAsmOutExt
data = gAttrFile + gFileOutExt
indexFile = open(index, "wt")
dataFile = open(data, "wb")
    
gapCount=0
for file in range(gMaxFiles):
    filename = str(file).zfill(3) + gFileInExt
    if os.path.isfile(filename):
        for g in range(gapCount):
            # Does not work - need to rewrite if any gaps between files.
            gAttBank.append([])
        # Get raw data.
        ScrFile = open(filename, "rb")
        ScrData = ScrFile.read()
        ScrFile.close()
        # Extract and write pixels.
        PixelsOut = open(str(file).zfill(3) + gFileOutExt, "wb")
        PixelData = bytearray(ScrData[:gScrAttStart])
        PixelsToBmp(PixelData, str(file).zfill(3) + gImgOutExt)
        PixelsOut.write(PixelData)
        PixelsOut.close()
        # Extract, compress, index, and write attributes.
        indexFile.write("  DW " + str(gAttrOffset) + "\n")
        AttrData=CompressAttributes(bytearray(ScrData[gScrAttStart:gScrLength]))
        dataFile.write(AttrData)
        gAttrOffset+=len(AttrData)
        gapCount=0
    else:
        gapCount+=1

# Done.    
indexFile.close()
dataFile.close()
sys.exit(0)
